@@ -4,12 +4,11 @@ import datetime
 from itertools import chain
 from decimal import Decimal
 
-from sql import Table, Literal, Null
-from sql.functions import Overlay, Position
+from sql import Literal, Null
 from sql.aggregate import Count
 from sql.operators import Concat
 
-from trytond.model import Workflow, ModelView, ModelSQL, fields, \
+from trytond.model import Workflow, Model, ModelView, ModelSQL, fields, \
     sequence_ordered
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
@@ -26,7 +25,8 @@ __all__ = ['Purchase', 'PurchaseIgnoredInvoice',
     'PurchaseRecreadtedInvoice', 'PurchaseLine', 'PurchaseLineTax',
     'PurchaseLineIgnoredMove', 'PurchaseLineRecreatedMove', 'PurchaseReport',
     'OpenSupplier', 'HandleShipmentExceptionAsk', 'HandleShipmentException',
-    'HandleInvoiceExceptionAsk', 'HandleInvoiceException']
+    'HandleInvoiceExceptionAsk', 'HandleInvoiceException', 'ModifyHeader',
+    'get_shipments_returns', 'search_shipments_returns']
 
 _STATES = {
     'readonly': Eval('state') != 'draft',
@@ -43,6 +43,36 @@ STATES = [
     ]
 
 
+def get_shipments_returns(model_name):
+    "Computes the returns or shipments"
+    def method(self, name):
+        Model = Pool().get(model_name)
+        shipments = set()
+        for line in self.lines:
+            for move in line.moves:
+                if isinstance(move.shipment, Model):
+                    shipments.add(move.shipment.id)
+        return list(shipments)
+    return method
+
+
+def search_shipments_returns(model_name):
+    "Search on shipments or returns"
+    def method(self, name, clause):
+        nested = clause[0].lstrip(name)
+        if nested:
+            return [('lines.moves.shipment' + nested,)
+                + tuple(clause[1:3]) + (model_name,) + tuple(clause[3:])]
+        else:
+            if isinstance(clause[2], str):
+                target = 'rec_name'
+            else:
+                target = 'id'
+            return [('lines.moves.shipment.' + target,)
+                + tuple(clause[1:3]) + (model_name,)]
+    return classmethod(method)
+
+
 class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     'Purchase'
     __name__ = 'purchase.purchase'
@@ -154,10 +184,10 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             ('exception', 'Exception'),
             ], 'Shipment State', readonly=True, required=True)
     shipments = fields.Function(fields.One2Many('stock.shipment.in', None,
-            'Shipments'), 'get_shipments')
+            'Shipments'), 'get_shipments', searcher='search_shipments')
     shipment_returns = fields.Function(
         fields.One2Many('stock.shipment.in.return', None, 'Shipment Returns'),
-        'get_shipment_returns')
+        'get_shipment_returns', searcher='search_shipment_returns')
     moves = fields.One2Many('stock.move', 'purchase', 'Moves', readonly=True)
 
     @classmethod
@@ -177,6 +207,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                 ('draft', 'quotation'),
                 ('quotation', 'confirmed'),
                 ('confirmed', 'processing'),
+                ('confirmed', 'draft'),
                 ('processing', 'processing'),
                 ('processing', 'done'),
                 ('done', 'processing'),
@@ -188,11 +219,14 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         cls._buttons.update({
                 'cancel': {
                     'invisible': ~Eval('state').in_(['draft', 'quotation']),
+                    'depends': ['state'],
                     },
                 'draft': {
-                    'invisible': ~Eval('state').in_(['cancel', 'quotation']),
-                    'icon': If(Eval('state') == 'cancel', 'tryton-clear',
-                        'tryton-go-previous'),
+                    'invisible': ~Eval('state').in_(
+                        ['cancel', 'quotation', 'confirmed']),
+                    'icon': If(Eval('state') == 'cancel', 'tryton-undo',
+                        'tryton-back'),
+                    'depends': ['state'],
                     },
                 'quote': {
                     'pre_validate': [
@@ -205,20 +239,30 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                         ],
                     'invisible': Eval('state') != 'draft',
                     'readonly': ~Eval('lines', []),
+                    'depends': ['state'],
                     },
                 'confirm': {
                     'invisible': Eval('state') != 'quotation',
+                    'depends': ['state'],
                     },
                 'process': {
                     'invisible': Eval('state') != 'confirmed',
+                    'depends': ['state'],
                     },
                 'handle_invoice_exception': {
                     'invisible': ((Eval('invoice_state') != 'exception')
                         | (Eval('state') == 'cancel')),
+                    'depends': ['state', 'invoice_state'],
                     },
                 'handle_shipment_exception': {
                     'invisible': ((Eval('shipment_state') != 'exception')
                         | (Eval('state') == 'cancel')),
+                    'depends': ['state', 'shipment_state'],
+                    },
+                'modify_header': {
+                    'invisible': ((Eval('state') != 'draft')
+                        | ~Eval('lines', [-1])),
+                    'depends': ['state'],
                     },
                 })
         # The states where amounts are cached
@@ -232,34 +276,9 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         InvoiceLine = pool.get('account.invoice.line')
         TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
-        model_data = Table('ir_model_data')
-        model_field = Table('ir_model_field')
         sql_table = cls.__table__()
 
-        # Migration from 1.2: packing renamed into shipment
-        cursor.execute(*model_data.update(
-                columns=[model_data.fs_id],
-                values=[Overlay(model_data.fs_id, 'shipment',
-                        Position('packing', model_data.fs_id),
-                        len('packing'))],
-                where=model_data.fs_id.like('%packing%')
-                & (model_data.module == module_name)))
-        cursor.execute(*model_field.update(
-                columns=[model_field.relation],
-                values=[Overlay(model_field.relation, 'shipment',
-                        Position('packing', model_field.relation),
-                        len('packing'))],
-                where=model_field.relation.like('%packing%')
-                & (model_field.module == module_name)))
-        cursor.execute(*model_field.update(
-                columns=[model_field.name],
-                values=[Overlay(model_field.name, 'shipment',
-                        Position('packing', model_field.name),
-                        len('packing'))],
-                where=model_field.name.like('%packing%')
-                & (model_field.module == module_name)))
-        table = TableHandler(cls, module_name)
-        table.column_rename('packing_state', 'shipment_state')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.8:
         if table.column_exist('supplier_reference'):
@@ -267,20 +286,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             table.column_rename('supplier_reference', 'reference')
 
         super(Purchase, cls).__register__(module_name)
-
-        # Migration from 1.2: rename packing to shipment in
-        # invoice_method values
-        cursor.execute(*sql_table.update(
-                columns=[sql_table.invoice_method],
-                values=['shipment'],
-                where=sql_table.invoice_method == 'packing'))
-
-        table = TableHandler(cls, module_name)
-        # Migration from 2.2: warehouse is no more required
-        table.not_null_action('warehouse', 'remove')
-
-        # Migration from 2.2: purchase_date is no more required
-        table.not_null_action('purchase_date', 'remove')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.2
         # state confirmed splitted into confirmed and processing
@@ -311,7 +317,6 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     where=sql_table.id.in_(sub_query.select(sub_query.id))))
 
         # Add index on create_date
-        table = TableHandler(cls, module_name)
         table.index_action('create_date', action='add')
 
     @classmethod
@@ -428,7 +433,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             for line in self.lines:
                 self.untaxed_amount += getattr(line, 'amount', None) or 0
             taxes = self._get_taxes()
-            self.tax_amount = sum(t['amount'] for t in taxes.itervalues())
+            self.tax_amount = sum(t['amount'] for t in taxes.values())
         if self.currency:
             self.untaxed_amount = self.currency.round(self.untaxed_amount)
             self.tax_amount = self.currency.round(self.tax_amount)
@@ -455,7 +460,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         return taxable_lines
 
     def get_tax_amount(self):
-        taxes = self._get_taxes().itervalues()
+        taxes = iter(self._get_taxes().values())
         return sum((tax['amount'] for tax in taxes), Decimal(0))
 
     @classmethod
@@ -495,7 +500,7 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
             'tax_amount': tax_amount,
             'total_amount': total_amount,
             }
-        for key in result.keys():
+        for key in list(result.keys()):
             if key not in names:
                 del result[key]
         return result
@@ -539,21 +544,13 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
                     'invoice_state': state,
                     })
 
-    def get_shipments_returns(model_name):
-        "Computes the returns or shipments"
-        def method(self, name):
-            Model = Pool().get(model_name)
-            shipments = set()
-            for line in self.lines:
-                for move in line.moves:
-                    if isinstance(move.shipment, Model):
-                        shipments.add(move.shipment.id)
-            return list(shipments)
-        return method
-
     get_shipments = get_shipments_returns('stock.shipment.in')
     get_shipment_returns = get_shipments_returns('stock.shipment.in.return')
 
+    search_shipments = search_shipments_returns('stock.shipment.in')
+    search_shipment_returns = search_shipments_returns(
+        'stock.shipment.in.return')
+
     def get_shipment_state(self):
         '''
         Return the shipment state for the purchase.
@@ -591,11 +588,14 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         return ''
 
     def get_rec_name(self, name):
+        items = []
         if self.number:
-            return self.number
-        elif self.reference:
-            return '[%s]' % self.reference
-        return '(%s)' % self.id
+            items.append(self.number)
+        if self.reference:
+            items.append('[%s]' % self.reference)
+        if not items:
+            items.append('(%s)' % self.id)
+        return ' '.join(items)
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -612,19 +612,31 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
 
     @classmethod
     def view_attributes(cls):
-        return [('/form//field[@name="comment"]', 'spell', Eval('party_lang'))]
+        attributes = [
+            ('/form//field[@name="comment"]', 'spell', Eval('party_lang'))]
+        if Transaction().context.get('modify_header'):
+            attributes.extend([
+                    ('//group[@id="states"]', 'states', {'invisible': True}),
+                    ('//group[@id="amount_buttons"]',
+                        'states', {'invisible': True}),
+                    ('//page[@name="invoices"]',
+                        'states', {'invisible': True}),
+                    ('//page[@name="shipments"]',
+                        'states', {'invisible': True}),
+                    ])
+        return attributes
 
     @classmethod
     def copy(cls, purchases, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['state'] = 'draft'
-        default['number'] = None
-        default['invoice_state'] = 'none'
-        default['invoices_ignored'] = None
-        default['moves'] = None
-        default['shipment_state'] = 'none'
+        else:
+            default = default.copy()
+        default.setdefault('number', None)
+        default.setdefault('invoice_state', 'none')
+        default.setdefault('invoices_ignored', None)
+        default.setdefault('moves', None)
+        default.setdefault('shipment_state', 'none')
         default.setdefault('purchase_date', None)
         return super(Purchase, cls).copy(purchases, default=default)
 
@@ -663,11 +675,10 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     @classmethod
     def store_cache(cls, purchases):
         for purchase in purchases:
-            cls.write([purchase], {
-                    'untaxed_amount_cache': purchase.untaxed_amount,
-                    'tax_amount_cache': purchase.tax_amount,
-                    'total_amount_cache': purchase.total_amount,
-                    })
+            purchase.untaxed_amount_cache = purchase.untaxed_amount
+            purchase.tax_amount_cache = purchase.tax_amount
+            purchase.total_amount_cache = purchase.total_amount
+        cls.save(purchases)
 
     def _get_invoice_purchase(self):
         'Return invoice'
@@ -801,8 +812,15 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     @ModelView.button
     @Workflow.transition('confirmed')
     def confirm(cls, purchases):
+        pool = Pool()
+        Configuration = pool.get('purchase.configuration')
         cls.set_purchase_date(purchases)
         cls.store_cache(purchases)
+        config = Configuration(1)
+        with Transaction().set_context(
+                queue_name='purchase',
+                queue_scheduled_at=config.purchase_process_after):
+            cls.__queue__.process(purchases)
 
     @classmethod
     @Workflow.transition('processing')
@@ -824,13 +842,6 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
     def handle_shipment_exception(cls, purchases):
         pass
 
-    @classmethod
-    def cron_process(cls):
-        purchases = cls.search([
-                ('state', '=', 'confirmed'),
-                ])
-        cls.process(purchases)
-
     @classmethod
     @ModelView.button
     def process(cls, purchases):
@@ -855,6 +866,11 @@ class Purchase(Workflow, ModelSQL, ModelView, TaxableMixin):
         if done:
             cls.do(done)
 
+    @classmethod
+    @ModelView.button_action('purchase.wizard_modify_header')
+    def modify_header(cls, purchases):
+        pass
+
 
 class PurchaseIgnoredInvoice(ModelSQL):
     'Purchase - Ignored Invoice'
@@ -879,7 +895,6 @@ class PurchaseRecreadtedInvoice(ModelSQL):
 class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     'Purchase Line'
     __name__ = 'purchase.line'
-    _rec_name = 'description'
     purchase = fields.Many2One('purchase.purchase', 'Purchase',
         ondelete='CASCADE', select=True, required=True,
         states={
@@ -921,10 +936,17 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
         'on_change_with_unit_digits')
     product = fields.Many2One('product.product', 'Product',
-        ondelete='RESTRICT', domain=[('purchasable', '=', True)],
+        ondelete='RESTRICT',
+        domain=[
+            ('purchasable', '=', True),
+            If(Bool(Eval('product_supplier')),
+                ('product_suppliers', '=', Eval('product_supplier')),
+                ()),
+            ],
         states={
             'invisible': Eval('type') != 'line',
             'readonly': Eval('purchase_state') != 'draft',
+            'required': Bool(Eval('product_supplier')),
             },
         context={
             'locations': If(Bool(Eval('_parent_purchase', {}).get(
@@ -941,7 +963,21 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
             'uom': Eval('unit'),
             'taxes': Eval('taxes', []),
             'quantity': Eval('quantity'),
-            }, depends=['type', 'purchase_state'])
+            }, depends=['type', 'purchase_state', 'product_supplier'])
+    product_supplier = fields.Many2One(
+        'purchase.product_supplier', "Supplier's Product",
+        ondelete='RESTRICT',
+        domain=[
+            If(Bool(Eval('product')),
+                ('product.products', '=', Eval('product')),
+                ()),
+            ('party', '=', Eval('_parent_purchase', {}).get('party')),
+            ],
+        states={
+            'invisible': Eval('type') != 'line',
+            'readonly': Eval('purchase_state') != 'draft',
+            },
+        depends=['product', 'type', 'purchase_state'])
     product_uom_category = fields.Function(
         fields.Many2One('product.uom.category', 'Product Uom Category'),
         'on_change_with_product_uom_category')
@@ -958,7 +994,7 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
                 'invisible': ~Eval('type').in_(['line', 'subtotal']),
                 },
             depends=['type']), 'get_amount')
-    description = fields.Text('Description', size=None, required=True,
+    description = fields.Text('Description', size=None,
         states={
             'readonly': Eval('purchase_state') != 'draft',
             },
@@ -1030,25 +1066,17 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
                     'purchase %(purchase)s misses an expense account.'),
                 'missing_default_account_expense': ('Purchase "%(purchase)s" '
                     'misses a default "account expense".'),
+                'delete_cancel_draft': ('The line "%(line)s" must be on '
+                    'canceled or draft purchase to be deleted.'),
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-        cursor = Transaction().connection.cursor()
-        sql_table = cls.__table__()
         super(PurchaseLine, cls).__register__(module_name)
-        table = TableHandler(cls, module_name)
+        table = cls.__table_handler__(module_name)
 
-        # Migration from 1.0 comment change into note
-        if table.column_exist('comment'):
-            cursor.execute(*sql_table.update(
-                    columns=[sql_table.note],
-                    values=[sql_table.comment]))
-            table.drop_column('comment', exception=True)
-
-        # Migration from 2.4: drop required on sequence
-        table.not_null_action('sequence', action='remove')
+        # Migration from 4.6: drop required on description
+        table.not_null_action('description', action='remove')
 
     @staticmethod
     def default_type():
@@ -1102,37 +1130,39 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         '''
         return {}
 
+    @fields.depends('purchase', '_parent_purchase.currency',
+        '_parent_purchase.party', '_parent_purchase.purchase_date',
+        'unit', 'product', 'product_supplier', 'taxes')
     def _get_context_purchase_price(self):
         context = {}
-        if getattr(self, 'purchase', None):
-            if getattr(self.purchase, 'currency', None):
+        if self.purchase:
+            if self.purchase.currency:
                 context['currency'] = self.purchase.currency.id
-            if getattr(self.purchase, 'party', None):
+            if self.purchase.party:
                 context['supplier'] = self.purchase.party.id
-            if getattr(self.purchase, 'purchase_date', None):
-                context['purchase_date'] = self.purchase.purchase_date
+            context['purchase_date'] = self.purchase.purchase_date
         if self.unit:
             context['uom'] = self.unit.id
-        else:
+        elif self.product:
             context['uom'] = self.product.purchase_uom.id
-        context['taxes'] = [t.id for t in self.taxes]
+        if self.product_supplier:
+            context['product_supplier'] = self.product_supplier.id
+        context['taxes'] = [t.id for t in self.taxes or []]
         return context
 
-    @fields.depends('product', 'unit', 'quantity', 'description', 'purchase',
-        '_parent_purchase.party', '_parent_purchase.currency',
-        '_parent_purchase.purchase_date')
+    @fields.depends('product', 'unit', 'quantity', 'purchase',
+        '_parent_purchase.party', 'product_supplier',
+        methods=['_get_tax_rule_pattern', '_get_context_purchase_price'])
     def on_change_product(self):
-        Product = Pool().get('product.product')
+        pool = Pool()
+        Product = pool.get('product.product')
 
         if not self.product:
             return
 
-        context = {}
         party = None
         if self.purchase and self.purchase.party:
             party = self.purchase.party
-            if party.lang:
-                context['language'] = party.lang.code
 
         # Set taxes before unit_price to have taxes in context of purchase
         # price
@@ -1156,6 +1186,16 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
             self.unit = self.product.purchase_uom
             self.unit_digits = self.product.purchase_uom.digits
 
+        if self.purchase and self.purchase.party:
+            product_suppliers = [ps for ps in self.product.product_suppliers
+                if ps.party == self.purchase.party]
+            if len(product_suppliers) == 1:
+                self.product_supplier, = product_suppliers
+        if (self.product_supplier
+                and (self.product_supplier
+                    not in self.product.product_suppliers)):
+            self.product_supplier = None
+
         with Transaction().set_context(self._get_context_purchase_price()):
             self.unit_price = Product.get_purchase_price([self.product],
                 abs(self.quantity or 0))[self.product.id]
@@ -1163,21 +1203,24 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
                 self.unit_price = self.unit_price.quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
 
-        if not self.description:
-            with Transaction().set_context(context):
-                self.description = Product(self.product.id).rec_name
-
         self.type = 'line'
         self.amount = self.on_change_with_amount()
 
+    @fields.depends('product', 'product_supplier',
+        methods=['on_change_product'])
+    def on_change_product_supplier(self):
+        if not self.product and self.product_supplier:
+            if len(self.product_supplier.product.products) == 1:
+                self.product, = self.product_supplier.product.products
+        self.on_change_product()
+
     @fields.depends('product')
     def on_change_with_product_uom_category(self, name=None):
         if self.product:
             return self.product.default_uom_category.id
 
-    @fields.depends('product', 'quantity', 'unit', 'taxes', 'purchase',
-        '_parent_purchase.currency', '_parent_purchase.party',
-        '_parent_purchase.purchase_date')
+    @fields.depends('product', 'quantity', 'unit',
+        methods=['_get_context_purchase_price'])
     def on_change_quantity(self):
         Product = Pool().get('product.product')
 
@@ -1191,11 +1234,11 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
                 self.unit_price = self.unit_price.quantize(
                     Decimal(1) / 10 ** self.__class__.unit_price.digits[1])
 
-    @fields.depends(methods=['quantity'])
+    @fields.depends(methods=['on_change_quantity'])
     def on_change_unit(self):
         self.on_change_quantity()
 
-    @fields.depends(methods=['quantity'])
+    @fields.depends(methods=['on_change_quantity'])
     def on_change_taxes(self):
         self.on_change_quantity()
 
@@ -1245,19 +1288,19 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
         'delivery_date_edit', 'delivery_date_store',
         '_parent_purchase.delivery_date')
     def on_change_with_delivery_date(self, name=None):
+        pool = Pool()
+        Date = pool.get('ir.date')
         if self.moves:
             dates = filter(
                 None, (m.effective_date or m.planned_date for m in self.moves
                     if m.state != 'cancel'))
-            if dates:
-                return min(dates)
-            else:
-                return
+            return min(dates, default=None)
+        delivery_date = None
         if self.delivery_date_edit:
-            return self.delivery_date_store
-        if self.purchase and self.purchase.delivery_date:
-            return self.purchase.delivery_date
-        if (self.product
+            delivery_date = self.delivery_date_store
+        elif self.purchase and self.purchase.delivery_date:
+            delivery_date = self.purchase.delivery_date
+        elif (self.product
                 and self.quantity is not None
                 and self.quantity > 0
                 and self.purchase and self.purchase.party
@@ -1268,8 +1311,11 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
                     delivery_date = product_supplier.compute_supply_date(
                         date=date)
                     if delivery_date == datetime.date.max:
-                        return None
-                    return delivery_date
+                        delivery_date = None
+                    break
+        if delivery_date and delivery_date < Date.today():
+            delivery_date = None
+        return delivery_date
 
     @fields.depends('purchase', '_parent_purchase.state')
     def on_change_with_purchase_state(self, name=None):
@@ -1317,11 +1363,8 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
                         'purchase': self.purchase.rec_name,
                         })
         else:
-            for name in ['default_product_account_expense',
-                    'default_category_account_expense']:
-                invoice_line.account = account_config.get_multivalue(name)
-                if invoice_line.account:
-                    break
+            invoice_line.account = account_config.get_multivalue(
+                'default_category_account_expense')
             if not invoice_line.account:
                 self.raise_user_error('missing_default_account_expense',
                     {'purchase': self.purchase.rec_name})
@@ -1446,7 +1489,23 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
 
     def _get_move_invoice_lines(self, move_type):
         'Return the invoice lines that should be shipped'
-        return [l for l in self.invoice_lines if not l.stock_moves]
+        if self.purchase.invoice_method == 'order':
+            return [l for l in self.invoice_lines]
+        else:
+            return [l for l in self.invoice_lines if not l.stock_moves]
+
+    def get_rec_name(self, name):
+        if self.product:
+            return '%s @ %s' % (self.product.rec_name, self.purchase.rec_name)
+        else:
+            return self.purchase.rec_name
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        return ['OR',
+            ('purchase.rec_name',) + tuple(clause[1:]),
+            ('product.rec_name',) + tuple(clause[1:]),
+            ]
 
     @classmethod
     def view_attributes(cls):
@@ -1458,15 +1517,25 @@ class PurchaseLine(sequence_ordered(), ModelSQL, ModelView):
                     }),
             ]
 
+    @classmethod
+    def delete(cls, lines):
+        for line in lines:
+            if line.purchase_state not in {'cancel', 'draft'}:
+                cls.raise_user_error('delete_cancel_draft', {
+                        'line': line.rec_name,
+                        })
+        super(PurchaseLine, cls).delete(lines)
+
     @classmethod
     def copy(cls, lines, default=None):
         if default is None:
             default = {}
-        default = default.copy()
-        default['moves'] = None
-        default['moves_ignored'] = None
-        default['moves_recreated'] = None
-        default['invoice_lines'] = None
+        else:
+            default = default.copy()
+        default.setdefault('moves', None)
+        default.setdefault('moves_ignored', None)
+        default.setdefault('moves_recreated', None)
+        default.setdefault('invoice_lines', None)
         return super(PurchaseLine, cls).copy(lines, default=default)
 
 
@@ -1509,6 +1578,14 @@ class PurchaseReport(CompanyReport):
         with Transaction().set_context(address_with_party=True):
             return super(PurchaseReport, cls).execute(ids, data)
 
+    @classmethod
+    def get_context(cls, records, data):
+        pool = Pool()
+        Date = pool.get('ir.date')
+        context = super(PurchaseReport, cls).get_context(records, data)
+        context['today'] = Date.today()
+        return context
+
 
 class OpenSupplier(Wizard):
     'Open Suppliers'
@@ -1545,20 +1622,6 @@ class HandleShipmentExceptionAsk(ModelView):
     domain_moves = fields.Many2Many(
         'stock.move', None, None, 'Domain Moves')
 
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        model = Table('ir_model')
-        # Migration from 1.2: packing renamed into shipment
-        cursor.execute(*model.update(
-                columns=[model.model],
-                values=[Overlay(model.model, 'shipment',
-                        Position('packing', model.model),
-                        len('packing'))],
-                where=model.model.like('%packing%')
-                & (model.module == module_name)))
-        super(HandleShipmentExceptionAsk, cls).__register__(module_name)
-
 
 class HandleShipmentException(Wizard):
     'Handle Shipment Exception'
@@ -1614,7 +1677,7 @@ class HandleShipmentException(Wizard):
                 'moves_recreated': [('add', moves_recreated)],
                 })
 
-        Purchase.process([purchase])
+        Purchase.__queue__.process([purchase])
         return 'end'
 
 
@@ -1675,5 +1738,76 @@ class HandleInvoiceException(Wizard):
             'invoices_recreated': [('add', invoices_recreated)],
             })
 
-        Purchase.process([purchase])
+        Purchase.__queue__.process([purchase])
+        return 'end'
+
+
+class ModifyHeaderStateView(StateView):
+    def get_view(self, wizard, state_name):
+        with Transaction().set_context(modify_header=True):
+            return super(ModifyHeaderStateView, self).get_view(
+                wizard, state_name)
+
+
+class ModifyHeader(Wizard):
+    "Modify Header"
+    __name__ = 'purchase.modify_header'
+    start = ModifyHeaderStateView('purchase.purchase',
+        'purchase.modify_header_form', [
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Modify", 'modify', 'tryton-ok', default=True),
+            ])
+    modify = StateTransition()
+
+    @classmethod
+    def __setup__(cls):
+        super(ModifyHeader, cls).__setup__()
+        cls._error_messages.update({
+                'not_in_draft': (
+                    'The purchase "%(purchase)s" must be in draft '
+                    'to modify header.'),
+                })
+
+    def get_purchase(self):
+        pool = Pool()
+        Purchase = pool.get('purchase.purchase')
+
+        purchase = Purchase(Transaction().context['active_id'])
+        if purchase.state != 'draft':
+            self.raise_user_error('not_in_draft', {
+                    'purchase': purchase.rec_name,
+                    })
+        return purchase
+
+    def default_start(self, fields):
+        purchase = self.get_purchase()
+        defaults = {}
+        for fieldname in fields:
+            value = getattr(purchase, fieldname)
+            if isinstance(value, Model):
+                if getattr(purchase.__class__, fieldname)._type == 'reference':
+                    value = str(value)
+                else:
+                    value = value.id
+            elif isinstance(value, (list, tuple)):
+                value = [r.id for r in value]
+            defaults[fieldname] = value
+
+        # Mimic an empty sale in draft state to get the fields' states right
+        defaults['lines'] = []
+        return defaults
+
+    def transition_modify(self):
+        pool = Pool()
+        Line = pool.get('purchase.line')
+
+        purchase = self.get_purchase()
+        purchase.__class__.write([purchase], self.start._save_values)
+
+        # Call on_change after the save to ensure parent sale
+        # has the modified values
+        for line in purchase.lines:
+            line.on_change_product()
+        Line.save(purchase.lines)
+
         return 'end'
